/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#ifdef __GNUG__
#pragma interface
#endif

template <class ACTION> class PATM_Automate {

  private :
  PATM_Etat** _etats ;
  short _prochainEtat ;
  ACTION** _actions ;
  short _prochainAction ;
  PATM_Etat* suivant(char) ;
  int insererEtat() ;
  int insererAction(ACTION* action) ;  

  public :
  PATM_Automate() : _etats(NULL), _prochainEtat(0), _actions(NULL),
  _prochainAction(0) { this->insererEtat() ; }
  ACTION* rechercher(char* cle) ;
  void insererCle(char* cle, ACTION* action) ;
  void dump() ;
} ;

template <class ACTION>
ACTION* PATM_Automate<ACTION>::rechercher(char* cle)
{
  PATM_Contexte ctx(cle) ;
  while (ctx.deplacer(this->_etats[ctx.etat])) ;
  if (ctx.fonction() > ACTION_INVALIDE) {
    return this->_actions[ctx.fonction()] ;
  } else {
    return NULL ;
  }
}

template <class ACTION>
int PATM_Automate<ACTION>::insererAction(ACTION* action)
{
  if (this->_actions == NULL) {
    this->_actions = new ACTION* ;
  } else {
    ACTION** old = this->_actions ;
    this->_actions = new ACTION* [_prochainAction+1] ;
    memcpy(this->_actions, old, sizeof(ACTION*)*_prochainAction) ;
    delete old ;
  }
  this->_actions[_prochainAction++] = action ;
  return _prochainAction - 1 ;
}

template <class ACTION>
int PATM_Automate<ACTION>::insererEtat()
{
  if (this->_etats == NULL) {
    this->_etats = new PATM_Etat* ;
  } else {
    PATM_Etat** old = this->_etats ;
    this->_etats = new PATM_Etat* [_prochainEtat+1] ;
    memcpy(this->_etats, old, sizeof(PATM_Etat*)*_prochainEtat) ;
    delete old ;
  }
  this->_etats[_prochainEtat++] = new PATM_Etat ;
  return _prochainEtat - 1 ;
}

template <class ACTION>
void PATM_Automate<ACTION>::insererCle(char* cle, ACTION* action)
{
  int etatnum = 0;
  int actnum = this->insererAction(action) ;
  int offs ;
  for (int i = 0; i < (strlen(cle) - 1); i++) {
    offs =  _prochainEtat - etatnum ;
    offs = this->_etats[etatnum]->insererTransition(cle[i], offs) ;
    if (offs < 0) {
      etatnum = this->insererEtat() ;
    } else {
      etatnum += offs ;
    }
  }
  this->_etats[etatnum]->insererTerminal(cle[i], actnum) ;
}

template <class ACTION>
void PATM_Automate<ACTION>::dump()
{
  int first=-1 ;
  int next=-1 ;
  char text[1024] ;

  text[0]= (char) 0 ;
  for (int i=0 ; i<_prochainEtat; i++) {
    this->_etats[i]->dump(i,first,next,text) ;
  }
}

